/*
 * #%L
 * Pollen :: Services
 * $Id$
 * $HeadURL$
 * %%
 * Copyright (C) 2009 - 2012 CodeLutin, Tony Chemit
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * #L%
 */
package org.chorem.pollen.services.impl;

import com.google.common.base.Charsets;
import com.google.common.collect.Lists;
import com.sun.syndication.feed.synd.SyndContent;
import com.sun.syndication.feed.synd.SyndContentImpl;
import com.sun.syndication.feed.synd.SyndEntry;
import com.sun.syndication.feed.synd.SyndEntryImpl;
import com.sun.syndication.feed.synd.SyndFeed;
import com.sun.syndication.feed.synd.SyndFeedImpl;
import com.sun.syndication.io.FeedException;
import com.sun.syndication.io.SyndFeedInput;
import com.sun.syndication.io.SyndFeedOutput;
import com.sun.syndication.io.XmlReader;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.chorem.pollen.PollenTechnicalException;
import org.chorem.pollen.bean.PollUrl;
import org.chorem.pollen.business.persistence.Choice;
import org.chorem.pollen.business.persistence.Comment;
import org.chorem.pollen.business.persistence.Poll;
import org.chorem.pollen.business.persistence.PollAccount;
import org.chorem.pollen.business.persistence.Vote;
import org.chorem.pollen.services.PollenServiceSupport;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.List;

import static org.nuiton.i18n.I18n._;
import static org.nuiton.i18n.I18n.l_;

/**
 * Service to manage feed of a poll.
 *
 * @author tchemit <chemit@codelutin.com>
 * @since 1.3
 */
public class PollFeedService extends PollenServiceSupport {

    /** Logger. */
    private static final Log log = LogFactory.getLog(PollFeedService.class);

    public boolean isFeedExists(Poll poll) {
        File file = getFeedLocation(poll);
        return file.exists();
    }

    /**
     * Obtain the feed file of the given poll.
     *
     * @param poll poll
     * @return the location of the feed of the given poll
     */
    public File getFeedLocation(Poll poll) {

        File feedDirectory = getConfiguration().getFeedDirectory();
        String filename = poll.getPollId() + ".xml";
        File result = new File(feedDirectory, filename);
        return result;
    }

    /**
     * Obtain the feed file of the given poll.
     *
     * @param poll poll
     * @return the location of the feed of the given poll
     */
    public String getFeed(Poll poll) {

        File feedDirectory = getConfiguration().getFeedDirectory();
        String filename = poll.getPollId() + ".xml";
        File feedFile = new File(feedDirectory, filename);
        try {
            String result = FileUtils.readFileToString(feedFile, Charsets.UTF_8);
            return result;
        } catch (IOException e) {
            throw new PollenTechnicalException("Could not read feed file", e);
        }
    }

    public void onPollCreated(Poll poll) {

        String title = _("pollen.feed.createPollContent",
                         poll.getCreator().getVotingId());

        String content = _("pollen.feed.createPollContent",
                           poll.getDescription());

        addFeedEntry(poll, title, content);

    }

    public void onCommentAdded(Poll poll, Comment comment) {

        String title = _("pollen.feed.addCommentTitle",
                         comment.getAuthor());

        String content = _("pollen.feed.addCommentContent",
                           comment.getText());

        addFeedEntry(poll, title, content);
    }

    public void onCommentDeleted(Poll poll, Comment comment, String reason) {

        String title = _("pollen.feed.deleteCommentTitle",
                         comment.getAuthor(), reason);

        String content =
                _("pollen.feed.deleteCommentContent", comment.getText());

        addFeedEntry(poll, title, content);
    }

    public void onVoteAdded(Poll poll, Vote vote) {

        PollVoteCountingService service =
                newService(PollVoteCountingService.class);

        String pollResult = service.getResultsAsString(poll);

        String userId;

        if (vote.isAnonymous()) {

            userId = _("pollen.common.anonymous");
        } else {
            userId = vote.getPollAccount().getVotingId();

        }
        String title = _("pollen.feed.addVoteTitle", userId);

        String content = _("pollen.feed.addVoteContent", pollResult);

        addFeedEntry(poll, title, content);
    }

    public void onVoteUpdated(Poll poll, Vote vote) {

        PollVoteCountingService service =
                newService(PollVoteCountingService.class);

        String pollResult = service.getResultsAsString(poll);

        String userId;

        if (vote.isAnonymous()) {

            userId = _("pollen.common.anonymous");
        } else {
            userId = vote.getPollAccount().getVotingId();

        }
        String title = _("pollen.feed.updateVoteTitle", userId);

        String content = _("pollen.feed.updateVoteContent", pollResult);

        addFeedEntry(poll, title, content);
    }

    public void onVoteDeleted(Poll poll, Vote vote,
                              PollAccount voteAccount, String reason) {

        PollVoteCountingService service =
                newService(PollVoteCountingService.class);

        String pollResult = service.getResultsAsString(poll);

        String userId;
        if (vote.isAnonymous()) {
            userId = _("pollen.common.anonymous");
        } else {
            userId = voteAccount.getVotingId();

        }
        String title = _("pollen.feed.deleteVoteTitle", userId, reason);

        String content = _("pollen.feed.deleteVoteContent", pollResult);

        addFeedEntry(poll, title, content);
    }

    public void onChoiceAdded(Poll poll, Choice choice) {

        String title = _("pollen.feed.addChoiceTitle", choice.getName());

        String content = _("pollen.feed.addChoiceContent",
                           choice.getDescription());

        addFeedEntry(poll, title, content);
    }

    /**
     * Delete the poll feed file.
     *
     * @param poll poll
     */
    public void deleteFeed(Poll poll) {

        File feedLocation = getFeedLocation(poll);
        if (feedLocation.exists()) {
            boolean wasDeleted = feedLocation.delete();
            if (!wasDeleted) {
                throw new PollenTechnicalException("Could not delete feed");
            }
            if (log.isDebugEnabled()) {
                log.debug("Feed deleted: " + feedLocation.getAbsolutePath());
            }
        }
    }

    /**
     * Ajout d'une entrée dans un flux.
     *
     * @param poll    poll
     * @param title   titre de l'entrée
     * @param content contenu de l'entrée
     */
    protected void addFeedEntry(Poll poll, String title, String content) {

        File file = getFeedLocation(poll);

        PollUrlService urlService = newService(PollUrlService.class);
        PollUrl voteUrl = urlService.getPollVoteUrl(poll);

        try {

            SyndFeed feed = getFeed(file, poll);

            List entries = Lists.newArrayList(feed.getEntries());

            SyndEntry entry;
            SyndContent description;
            entry = new SyndEntryImpl();
            entry.setTitle(title);
            entry.setLink(voteUrl.getUrl());
            entry.setPublishedDate(serviceContext.getCurrentTime());
            description = new SyndContentImpl();
            description.setType("text/plain");
            description.setValue(content);
            entry.setDescription(description);
            entries.add(entry);

            feed.setEntries(entries);

            Writer writer = new FileWriter(file);
            try {
                SyndFeedOutput output = new SyndFeedOutput();
                output.output(feed, writer);
                writer.close();
            } finally {
                IOUtils.closeQuietly(writer);
            }

            if (log.isDebugEnabled()) {
                log.debug("Feed updated: " + file.getAbsolutePath());
            }
        } catch (PollenTechnicalException e) {
            throw e;
        } catch (Exception e) {
            throw new PollenTechnicalException("Could not feed feed", e);
        }
    }

    protected SyndFeed getFeed(File file, Poll poll) throws IOException, FeedException {

        SyndFeedInput input = new SyndFeedInput();

        SyndFeed feed;

        if (!file.exists()) {

            // create the feed
            feed = createFeed(file, poll);

        } else {

            // load the feed

            try {
                feed = input.build(new XmlReader(file));
            } catch (FeedException e) {

                // create the feed
                feed = createFeed(file, poll);

            }

        }

        return feed;

    }

    protected SyndFeed createFeed(File file, Poll poll) throws IOException, FeedException {

        String title = l_(getLocale(), "pollen.feed.pollTitle",
                          poll.getTitle());
        String description = l_(getLocale(), "pollen.feed.pollDescription",
                                poll.getDescription());

        PollUrlService urlService = newService(PollUrlService.class);
        PollUrl voteUrl = urlService.getPollVoteUrl(poll);

        SyndFeed feed = new SyndFeedImpl();
        feed.setFeedType("atom_1.0");
        feed.setTitle(title);
        feed.setLink(voteUrl.getUrl());
        feed.setDescription(description);

        Writer writer = new FileWriter(file);

        try {

            SyndFeedOutput output = new SyndFeedOutput();
            output.output(feed, writer);
            writer.close();

        } finally {
            IOUtils.closeQuietly(writer);
        }

        if (log.isDebugEnabled()) {
            log.debug("Feed created: " + file.getAbsolutePath());
        }

        return feed;

    }

}
